Require Import Tools.
Require Import ZArith.
Open Scope Z_scope.
Require Import List.
Require Import Recdef.
Require Import Zwf.
Require Import Zdiv.

Section Decomp.

 Variable prime : Z -> Prop.
 Hypothesis prime_1 : forall n:Z, prime n -> 2 <= n.
 Hypothesis prime_2 : forall n p q:Z,  prime n -> n = p * q -> n = p \/ n = q.

 Variable least_prime :  Z -> Z.

 Hypothesis least_prime_1 : forall n, 2 <= n -> prime(least_prime n).
 Hypothesis least_prime_2 :
     forall n, 2 <= n -> let p := least_prime n 
                         in n = p*(n/p).
 Hypothesis least_prime_3 : forall n , 2 <= n -> forall p q, prime p ->
                                       n = p*q -> least_prime n <= p.

 (* question 1 :
    Define a function decomp : Z -> list Z such that
    if 2 <= n, then decomp n is a list of prime number whose product is
    equal no n 

    question 2: prove the correctness of your function *)

   


(* Answer *)


 Function decomp (n:Z){wf (Zwf 1) n}: list Z :=
  if Zle_bool n 1 then nil
  else let p := least_prime n 
       in if Zeq_bool n p then n::nil
          else p :: (decomp (n/p)).
Proof.
intros n teq teq0;Zbool2Prop.
split.
omega.
apply Z_div_lt.
SearchPattern (_ >= _).
apply Zle_ge.
auto with zarith.
auto with zarith.
apply Zwf_well_founded.
Defined.


Lemma decomp_all_primes : forall n, 
                     forall i, In i (decomp n) -> prime i.
Proof.
intros n;functional induction (decomp n).
inversion 1.
inversion 1.
subst i.
Zbool2Prop.
rewrite e0;auto.
auto with zarith.
destruct H0.
intros i Hi;destruct  (in_inv  Hi).
rewrite <- H.
Zbool2Prop.
auto with zarith.
auto.
Qed.


Lemma decomp_product : forall n, 1 <= n -> n = fold_right (Zmult) 1 (decomp n).
Proof.
  intro n;functional induction (decomp n);simpl;auto with zarith.
 Zbool2Prop; eauto with zarith.
 intro H;simpl.
 rewrite <- IHl.
rewrite <- least_prime_2;auto.
 Zbool2Prop;omega.
assert (2 <= n) by (Zbool2Prop; omega).

SearchPattern (_ <= _ \/ _ < _).
destruct (Zle_or_lt 1 (n / least_prime n));auto.
assert ( n / least_prime n <= 0).
omega.
assert (0 <= least_prime n).
 generalize (least_prime_1 n);intro H3.
apply H3 in H0.

apply prime_1 in H0.
omega.


generalize (Zmult_le_compat_r _ _   _  H2 H3).
rewrite Zmult_comm.
rewrite <- least_prime_2.
intro;rewrite Zmult_0_l in *;omega.
auto.
Qed.

End Decomp.



