(* Tools for using a list of measures *)



Require Import Wf.
Require Import Lexicographic_Product.
Require Import Inclusion.
Require Import Inverse_Image. 
Require Import Recdef.

(* For simplicity's sake, we first introduce a non-dependent version
   of Standard Library's lexicoqgraphic product *)


Section Nondep_Lexicographic_Product.
  Variable A : Type.
  Variable B : Type.
  Variable leA : A -> A -> Prop.
  Variable leB :  B -> B  -> Prop.

  Let lex0 :=  Relation_Operators.lexprod A (fun a => B)  leA
                                            (fun _  => leB).

 (* Non dependent lexicographic product *)

 Definition  lex : A * B -> A * B -> Prop :=
  fun p q : A * B =>
   lex0 (existT (fun _ : A => B) (fst p) (snd p))
        (existT (fun _ : A => B) (fst q) (snd q)).

Lemma lex_wf: 
    well_founded leA -> well_founded  leB -> 
    well_founded lex.
Proof.
 intros H H0;unfold lex. 
 generalize (wf_lexprod  _ _   leA  (fun _  => leB) H (fun a => H0)).
intros H1;apply  (wf_inverse_image).  
apply  wf_lexprod;auto.
Qed.
Print lex_wf.




End Nondep_Lexicographic_Product.
Implicit Arguments lex [A B].
Require Import Arith.
Require Import Recdef.



Require Import List.


Section Measures.

 Variable A : Type.


Require Import Compare_dec.


 Definition measure_order (f : A -> nat) : A -> A -> Prop :=
    fun x y =>  f x < f y.

About lex.

Fixpoint measures_lt (l : list (A -> nat)): A -> A -> Prop :=
  match l with nil => (fun a b => False)
             | f::l' => (fun a b => f a < f b \/ (f a = f b /\ measures_lt l' a b))
 end.


Lemma measures_lt_wf : forall l, well_founded  (measures_lt l).
induction l.
 intro x.
 split.
 inversion 1.
 pose (o1 := lex lt (measures_lt l)).
 assert (well_founded o1).
 apply lex_wf;auto.
 apply lt_wf.

 pose (o2 := fun x y : A => o1( a x, x) (a y,y)).
 assert (well_founded o2).
 unfold o2.
 apply wf_inverse_image.
 auto.
 eapply wf_incl.
 2:eexact H0.
 intros x y Hxy.
 simpl in Hxy.
 unfold o2,o1.
 destruct Hxy.
 left.
 simpl;auto.
 simpl. 
  destruct H1 as [H2 H3].
 rewrite H2.
 right;auto.
Qed.
 
End Measures.

About measures_lt.

Implicit Arguments measures_lt [A].
