Require Import ZArith.
Open Scope Z_scope.
Require Import Recdef.
Require Import Zwf.
Require Import Tools.

Open Scope Z_scope.
Section Definitions.

 Definition ptr_t := Z.
 Definition size_t := Z.

 Definition array_t := ptr_t  -> Z.
 Definition search_t := ptr_t -> size_t -> option ptr_t.
 Variable a : array_t. 
 Variable x : Z.

Definition correct (s : search_t) from n:=
  forall  i,
   s from n = Some i ->from <= i /\ i <= from+  n-1  /\ a(i)=x.

Definition complete (s : search_t) from n :=
  s from n = None ->
                               forall k, from <= k ->
                                         k < from + n ->
                                         a k <> x.




Definition middle (n p : Z) := (n + p)/ 2.

Lemma middle_ge : forall n p, n <= p -> n <= middle n p.
intros n p H.
pattern n at 1;replace n with ((n*2)/2).
unfold middle;apply Z_div_le;auto with zarith.
rewrite Z_div_mult_full;auto with zarith.
Qed.



Lemma middle_le : forall n p, n <= p -> middle n p <= p.
intros n p H.
pattern p at 2;replace p with ((p*2)/2).
unfold middle;apply Z_div_le;auto with zarith.
rewrite Z_div_mult_full;auto with zarith.
Qed.


Ltac abstract_middle n p name :=
   generalize (middle_le n p);generalize (middle_ge n p);
   pose (name := middle n p); fold name.

(* recherche de x entre from et from+n-1 *)

Function search (from  : ptr_t)(n : Z){wf (Zwf 0)  n} : 
option Z  :=
 if Zle_bool 1 n
 then
 let m := middle from (from +  n-1) in
  if Zeq_bool x (a m) 
  then Some m
  else if andb (Zle_bool (from+1)  m) (Zlt_bool x (a m))
       then search  from  (m - from) 
       else if (Zle_bool (m +1) (from + (n-1))) 
             then search  (m+1)  (from + n -1 -  m) 
              else None
 else None.
 intros; Zbool2Prop. 
abstract_middle from (from+n-1) m;split;try omega.
split;Zbool2Prop;try omega.
abstract_middle from (from+n-1) m;intros;omega.
intros;abstract_middle from (from+n-1) m; omega.
apply Zwf_well_founded.
Defined.


Lemma search_1 : forall from n, correct search from n. 
red;intros.
functional induction (search from n).
injection H;intro H0;subst i;clear H;Zbool2Prop.
abstract_middle from (from + n - 1) m.
fold m in e0 ;omega. 
split.
destruct IHo;auto.
split.
destruct (IHo H).
destruct H1.
Zbool2Prop;abstract_middle from (from + n - 1) m.
fold m in e0, IHo, H, H1, H3, H4;omega.
destruct (IHo H);tauto.

Zbool2Prop;abstract_middle from (from + n - 1) m.
fold m in e0,e1,e2,IHo,H.
intros;clear e0;Zbool2Prop.
assert (m < from +1) by omega.
destruct (IHo H);repeat (split;try omega).
destruct (IHo H);repeat (split;try omega).
fold m in H1, H0; omega.
discriminate.
discriminate.
Qed.


Definition sorted (from n : Z) := forall i j, from <= i ->
                                                i <= j ->
                                                j < from + n ->
                                                (a i <= a j)%Z.
Lemma sorted_restriction : 
   forall from n from' n', sorted from n ->from <= from' ->
                           from'+n' <= from+n 
                            -> sorted from' n'.
 unfold sorted;intros.
 apply H;omega.
Qed.


Lemma search_2 : forall from n, sorted from n -> complete search from  n. 
unfold complete;
intros from n H;functional induction (search from n).
intros;discriminate.
intros;abstract_middle from (from+n-1) m.
intros m_1 m_2;fold m in e, e0, IHo, H0.
Zbool2Prop.
destruct (Z_dec  k (middle from (from + n -1))).
destruct s.
apply IHo;auto.
apply sorted_restriction with from n;auto with zarith.
fold m in H3,H4,z;omega.
fold m in H3,H4,z;assert (a m <= a k).
apply H;try omega.
omega.
fold m in H3,H4,e1;subst k;omega.
intros;abstract_middle from (from+n-1) m;intros m_1 m_2.
fold m in e, e0, IHo, H0;Zbool2Prop.
destruct (Z_dec  m k ).
 destruct s.
assert (m = from).
fold m in e1,e2;omega.
fold m in e1,e2;assert (a m <= a k)%Z.
apply H ;auto;try omega.
apply IHo;auto;try omega.
red;intros;apply H;auto;try omega.
apply IHo;auto; try omega.
apply sorted_restriction with from n;auto.
omega.
omega.
fold m in e1,e2;omega.
subst k;omega.
destruct (Z_dec  m k ).
 destruct s.
(* cas m < k *)
apply IHo;auto;try omega.
apply sorted_restriction with from n;auto;try omega.

(* cas m > k *)
fold m in e1;assert (a m < x).
omega.
intro eg;subst x;clear H3;assert (a k <= a m)%Z.
apply H;auto;try omega.
omega.
subst k;omega.
intros;Zbool2Prop.
abstract_middle from (from+n-1) m;intros;fold m in  e1, e2.
assert (from = m).
unfold ptr_t in *;omega.
 rewrite H5 in e2;assert (n=1). 
 omega.
assert (k = from).
omega.

rewrite H7; rewrite H5;auto.
abstract_middle from (from+n-1) m;intros.
fold m in e0,e1, e2;assert (m = from + n -1).
omega.
assert (a m < x).
omega.
intro eg;subst x;assert (a k <= a m).
apply H;try omega.
omega.
intros;Zbool2Prop.
omega.
Qed.



End Definitions.
